const int PIN_RED = 11;
const int PIN_GRN = 10;
const int PIN_BLU = 9;

class RGBTransition {
  private:
    int pinR, pinG, pinB;
    int initH = 0, destH = 0;
    float initS = 0, destS = 0;
    float initV = 0, destV = 0;
    unsigned long startTime, transitionTime;
    bool inProgress = false;
    void (*onFinishListener)() = NULL;

    int sgn(int num) {
      if(num < 0) return -1;
      if(num > 0) return 1;
      return 0;
    }
    
  public:
    RGBTransition(int r, int g, int b) {
      pinR = r;
      pinG = g;
      pinB = b;
      pinMode(r, OUTPUT);
      pinMode(g, OUTPUT);
      pinMode(b, OUTPUT);
    }

    RGBTransition& setTransitionTime(unsigned long time) {
      transitionTime = time;
      return *this;
    }
    
    RGBTransition& setHue(int hue) {
      destH = hue;
      return *this;
    }

    RGBTransition& setSaturation(float sat) {
      destS = sat;
      return *this;
    }

    RGBTransition& setValue(float val) {
      destV = val;
      return *this;
    }

    RGBTransition& setInitHue(int hue) {
      initH = hue;
      return *this;
    }

    RGBTransition& setInitSaturation(float sat) {
      initS = sat;
      return *this;
    }

    RGBTransition& setInitValue(float val) {
      initV = val;
      return *this;
    }

    RGBTransition& setOnFinishListener(void (*listener)()) {
      onFinishListener = listener;
      return *this;
    }

    void doTransition(bool reset) {
      if(reset) {
        startTime = millis();
        inProgress = true;
      }
      float progress = (millis() - startTime) / (transitionTime * 1.0);
      if(inProgress && progress >= 1) {
        initH = destH;
        initS = destS;
        initV = destV;
        inProgress = false;
        if(onFinishListener != NULL) {
          (*onFinishListener)();
        }
      }
      
      if(!inProgress) return;

      int H = destH - initH;
      if(abs(H) < 180) {
        H = initH + (1.0 * H * progress);
      } else {
        H = initH + (-1.0 * sgn(H) * (360 - abs(H)) * progress);
        if(H < 0) H += 359;
      }

      float S = destS - initS;
      S = initS + (1.0 * S * progress);

      float V = destV - initV;
      V = initV + (1.0 * V * progress);
      
      setHSV(H, S, V);
    }

    void doTransition() {
      doTransition(false);
    }

    void setHSV(int H, float S, float V) {
      H %= 360;
      float C = V * S;
      float H2 = H/60.0 - (H/60) + ((H/60)%2);
      float X = C * (1.0 - abs(H2 - 1));
      float m = V - C;
      float R,G,B;
      if(H < 60) {
        R = C;
        G = X;
        B = 0;
      } else if(H < 120) {
        R = X;
        G = C;
        B = 0;
      } else if(H < 180) {
        R = 0;
        G = C;
        B = X;
      } else if(H < 240) {
        R = 0;
        G = X;
        B = C;
      } else if(H < 300) {
        R = X;
        G = 0;
        B = C;
      } else {
        R = C;
        G = 0;
        B = X;
      }
    
      R = (int)((R + m) * 255);
      G = (int)((G + m) * 255);
      B = (int)((B + m) * 255);
      
      analogWrite(pinR, R);
      analogWrite(pinG, G);
      analogWrite(pinB, B);
    }
};

RGBTransition led(PIN_RED, PIN_GRN, PIN_BLU);
int currentHue = 0;

void setup() {
  led.setTransitionTime(200);
  
  led.setInitHue(0).setHue(currentHue);
  led.setInitSaturation(1).setSaturation(1);
  led.setInitValue(0).setValue(1);
  
  led.setOnFinishListener(fadeToBlack);
  led.doTransition(true);
}

void loop() {
  led.doTransition();
}

void fadeToBlack() {
  led.setValue(0);
  led.setOnFinishListener(changeColor);
  led.doTransition(true);
}

void changeColor() {
  currentHue = (currentHue + 10) % 359;
  led.setHue(currentHue);
  led.setValue(1);
  led.setOnFinishListener(fadeToBlack);
  led.doTransition(true);
}
